# The current version of ACROPOLIS
version     = "1.3.0"

# The current dev version of ACROPOLIS
dev_version = "1.3.0"

# The short description of ACROPOLIS
description = "A generiC fRamework fOr Photodisintegration Of LIght elementS"

# The main webpage of ACROPOLIS
url         = "https://acropolis.hepforge.org"

# The authors of ACROPOLIS
authors     = "Paul Frederik Depta, Marco Hufnagel, Kai Schmidt-Hoberg"
